class ConnectionStatus:
    """Connection Status"""

    Connected = 'connected'
    Connecting = 'connecting'
    Disconnected = 'disconnected'


class Product:
    """Product"""

    Portal = 'Portal'
    Edge = 'Edge'
    Drive = 'Drive'
